from ._base import BaseWeights
import numpy as np
from scipy.stats import beta


class GeometricProcess(BaseWeights):
    def __init__(self, a=1, b=1, rng=None):
        super().__init__(rng=rng)
        self.a = a
        self.b = b
        self.p = self.rng.beta(a=self.a, b=self.b)

        self.v = np.array([], dtype=np.float64)

    def structure_log_likelihood(self, v=None, p=None, a=None, b=None):
        if v is None:
            v = self.v
        if p is None:
            p = self.p
        if a is None:
            a = self.a
        if b is None:
            b = self.b
        log_likelihood = self.weight_log_likelihood(v=v)
        log_likelihood += self.p_log_likelihood(p=p, a=a, b=b)
        return log_likelihood

    def weight_log_likelihood(self, v=None):
        if v is None:
            v = self.v
        if len(v) == 0:
            return 0
        if np.all(v == v[0]):
            return 0
        else:
            return -np.inf

    def p_log_likelihood(self, p=None, a=None, b=None):
        if p is None:
            p = self.p
        if a is None:
            a = self.a
        if b is None:
            b = self.b
        return beta.logpdf(p, a=a, b=b)

    def random(self, size=None):
        if size is None and len(self.d) == 0:
            raise ValueError("Weight structure not fitted and `n` not passed.")
        if size is None:
            size = max(self.d) + 1
        else:
            if type(size) is not int:
                raise TypeError("size parameter must be integer or None")
        self.p = self.rng.beta(self.a + len(self.d), self.b + self.d.sum())
        if size <= 0:
            size = 1
        self.v = np.repeat(self.p, size)
        self.w = self.v * np.cumprod(np.concatenate(([1],
                                                     1 - self.v[:-1])))
        return self.w

    def complete(self, size):
        if type(size) is not int:
            raise TypeError("size parameter must be integer or None")
        if len(self.w) < size:
            self.v = np.repeat(self.p, size)
            self.w = self.v * np.cumprod(np.concatenate(([1], 1 - self.v[:-1])))
        return self.w

    def tail(self, x):
        if x >= 1 or x < 0:
            raise ValueError("Tail parameter not in range [0,1)")
        size = int(np.log(1 - x) / np.log(1 - self.p))
        if size <= 0:
            size = 1
        self.v = np.repeat(self.p, size)
        self.w = self.v * np.cumprod(np.concatenate(([1], 1 - self.v[:-1])))
        return self.w

    def get_p(self):
        return self.p



from ._dirichlet_distribution import DirichletDistribution
from ._dirichlet_process import DirichletProcess
from ._pitman_yor_process import PitmanYorProcess
from ._geometric_process import GeometricProcess
from ._beta_in_beta import BetaInBeta
from ._beta_bernoulli import BetaBernoulli
from ._beta_binomial import BetaBinomial


"""Base class for weighting structure models."""
from abc import ABCMeta, abstractmethod
import numpy as np


class BaseWeights(metaclass=ABCMeta):
    """Base class for weighting structure models.

    This abstract class specifies an interface for all weighting structure
    classes and provides basic common methods for weighting models.
    """
    def __init__(self, rng=None):
        if rng is None:
            self.rng = np.random.default_rng()
        elif type(rng) is int:
            self.rng = np.random.default_rng(rng)
        else:
            self.rng = rng

        self.w = np.array([], dtype=np.float64)
        self.d = np.array([], dtype=np.int64)

    def fit(self, d):
        """Fit the weighting structure to a vector of assignments

        This method fits the parameters of the weighting model given the
        internal truncated weighting structure `self.w`.
        """
        self.d = np.array(d)

    @abstractmethod
    def random(self, size=None):
        """Do a random draw of the truncated weighting structure up to `n` obs.

        This method updates internal truncated weighting structure `self.w`
        as a step of the underlying Gibbs sampler.
        """
        pass

    @abstractmethod
    def tail(self, x):
        """Return an array of weights such that the sum is greater than `x`"""
        pass

    @abstractmethod
    def complete(self, size):
        """Return an array of weights with at least `n` elements"""
        pass

    @abstractmethod
    def structure_log_likelihood(self):
        """Return the log-likelihood of the complete weighting structure"""
        pass

    @abstractmethod
    def weight_log_likelihood(self, w=None):
        """Return the log-likelihood of the weights conditional to params"""
        pass

    def assign_log_likelihood(self, d=None):
        """Returns the log-likelihood of an assignment `d` given the weights"""
        if d is None:
            d = self.d
        self.complete(max(d))
        return np.sum(np.log(self.w[d]))

    def reset(self):
        """Resets the conditional vector `d` to None"""
        self.d = None

    def get_weights(self):
        """Returns the last weighting structure drawn"""
        return self.w

    def get_normalized_weights(self):
        """Returns the last weighting stricture normalized"""
        return self.w/np.sum(self.w)

    def get_normalized_cumulative_weights(self):
        """Returns the normalized cumulative weights"""
        return np.cumsum(self.get_normalized_weights())

    def get_size(self):
        """Returns the size of the truncated weighting structure"""
        return len(self.w)

    def random_assignment(self, size=None):
        """Returns a sample draw of the categorical assignment from the current
        state normalized weighting structure"""
        u = self.rng.uniform(size=size)
        self.tail(1-np.min(u))
        inverse_sampling = np.greater.outer(
            u, self.get_normalized_cumulative_weights()
        )
        return np.sum(inverse_sampling, axis=1)



import unittest
import numpy as np
import scipy.stats

from .. import mixture_models


class TestWeightModels(unittest.TestCase):
    def test_dirichlet_process(self):
        rng = np.random.default_rng(0)
        n = 100
        means = np.array([-10, 10])
        sds = np.array([1, 1])
        weights = np.array([0.5, 0.5])
        theta = rng.choice(range(len(weights)), size=n, p=weights)
        y = np.array([
            scipy.stats.multivariate_normal.rvs(means[j], sds[j],
                                                random_state=rng) for j in theta])

        dgp = mixture_models.BetaInBetaMixture(rng=rng)
        dgp.fit_gibbs(y)
        pass



from ._base import BaseWeights
import numpy as np
from scipy.stats import beta


class DirichletProcess(BaseWeights):
    def __init__(self, alpha=1, rng=None):
        super().__init__(rng=rng)
        self.alpha = alpha
        self.v = np.array([], dtype=np.float64)

    def structure_log_likelihood(self, v=None, alpha=None):
        if v is None:
            v = self.v
        if alpha is None:
            alpha = self.alpha
        return self.weight_log_likelihood(v=v, alpha=alpha)

    def weight_log_likelihood(self, v=None, alpha=None):
        if v is None:
            v = self.v
        if alpha is None:
            alpha = self.alpha
        if len(v) == 0:
            return 0
        return np.sum(beta.logpdf(v, a=1, b=alpha))

    def random(self, size=None):
        if size is None and len(self.d) == 0:
            raise ValueError("Weight structure not fitted and `n` not passed.")
        if size is not None:
            if type(size) is not int:
                raise TypeError("size parameter must be integer or None")
        if len(self.d) == 0:
            self.v = self.rng.beta(a=1, b=self.alpha, size=size)
            self.w = self.v * np.cumprod(np.concatenate(([1],
                                                         1 - self.v[:-1])))
        else:
            a_c = np.bincount(self.d)
            b_c = np.concatenate((np.cumsum(a_c[::-1])[-2::-1], [0]))

            if size is not None and size < len(a_c):
                a_c = a_c[:size]
                b_c = b_c[:size]

            self.v = self.rng.beta(a=1 + a_c, b=self.alpha + b_c)
            self.w = self.v * np.cumprod(np.concatenate(([1],
                                                         1 - self.v[:-1])))
            if size is not None:
                self.complete(size)
        return self.w

    def complete(self, size):
        if type(size) is not int:
            raise TypeError("size parameter must be integer or None")
        if len(self.v) < size:
            self.v = np.concatenate(
                (self.v,
                 self.rng.beta(a=1, b=self.alpha, size=size - len(self.v))))
            self.w = self.v * np.cumprod(np.concatenate(([1],
                                                         1 - self.v[:-1])))
        return self.w

    def tail(self, x):
        if x >= 1 or x < 0:
            raise ValueError("Tail parameter not in range [0,1)")
        if len(self.w) == 0:
            self.random(1)

        w_sum = sum(self.w)
        while w_sum < x:
            v_to_append = self.rng.beta(a=1, b=self.alpha, size=1)
            self.v = np.concatenate((self.v, v_to_append))
            self.w = np.concatenate((self.w, [(1 - sum(self.w)) * self.v[-1]]))
            w_sum += self.w[-1]
        return self.w



from ._base import BaseGaussianMixture
from ..weight_models import BetaInBeta


class BetaInBetaMixture(BaseGaussianMixture):
    def __init__(self, *, x=0.5, alpha=1, a=1, b=1, mu_prior=None,
                 lambda_prior=1, psi_prior=None, nu_prior=None,
                 p_method="static", total_iter=1000,
                 burn_in=100, subsample_steps=1, rng=None):
        weight_model = BetaInBeta(x=x, alpha=alpha, a=a, b=b,
                                  p_method=p_method, rng=rng)
        super().__init__(weight_model=weight_model, mu_prior=mu_prior,
                         lambda_prior=lambda_prior, psi_prior=psi_prior,
                         nu_prior=nu_prior, total_iter=total_iter,
                         burn_in=burn_in, subsample_steps=subsample_steps,
                         rng=rng)

    def _get_run_params(self):
        return {"w": self.weight_model.get_weights(),
                "mu": self.mu,
                "sigma": self.sigma,
                "u": self.u,
                "d": self.d,
                "p": self.weight_model.get_p()}



from ._base import BaseGaussianMixture
from ..weight_models import PitmanYorProcess


class PitmanYorMixture(BaseGaussianMixture):
    def __init__(self, *, pyd=0, theta=1, mu_prior=None,
                 lambda_prior=1, psi_prior=None, nu_prior=None, total_iter=1000,
                 burn_in=100, subsample_steps=1, rng=None):
        weight_model = PitmanYorProcess(pyd=pyd, alpha=theta, rng=rng)
        super().__init__(weight_model=weight_model, mu_prior=mu_prior,
                         lambda_prior=lambda_prior, psi_prior=psi_prior,
                         nu_prior=nu_prior, total_iter=total_iter,
                         burn_in=burn_in, subsample_steps=subsample_steps,
                         rng=rng)



from ._base import BaseGaussianMixture
from ..weight_models import DirichletProcess


class DirichletProcessMixture(BaseGaussianMixture):
    def __init__(self, *, theta=1, rng=None, **kwargs):
        weight_model = DirichletProcess(alpha=theta, rng=rng)
        super().__init__(weight_model=weight_model, rng=rng, **kwargs)



from ._base import BaseGaussianMixture
from ..weight_models import BetaBernoulli

import numpy as np


class BetaBernoulliMixture(BaseGaussianMixture):
    def __init__(self, *, p=0.5, theta=1, rng=None, **kwargs):
        weight_model = BetaBernoulli(p=p, alpha=theta, rng=rng)
        super().__init__(weight_model=weight_model, rng=rng, **kwargs)

    def _get_run_params(self):
        return {"w": self.weight_model.get_weights(),
                "mu": self.mu,
                "sigma": self.sigma,
                "u": self.u,
                "d": self.d,
                "p": self.weight_model.get_p()}



from scipy.stats import invwishart, multivariate_normal
from itertools import repeat
import numpy as np


def random_normal_invw(mu, lam, psi, nu, rng=None):
    if rng is None:
        rng = np.random.default_rng()
    ret_sigma = invwishart.rvs(nu, psi,
                               random_state=rng)
    ret_mu = multivariate_normal.rvs(mu, ret_sigma / lam,
                                     random_state=rng)
    return ret_mu, ret_sigma


def posterior_norm_invw_params(y, mu, lam, psi, nu):
    n, _ = y.shape
    ret_mu = (lam * mu + n * y.mean(axis=0)) / (lam + n)
    ret_lam = lam + n
    ret_psi = psi + n * np.cov(y.T, bias=True) + (
            (lam * n) / (lam + n) *
            ((y.mean(axis=0) - mu) @ (y.mean(axis=0) - mu)))
    ret_nu = nu + n
    return {"mu": ret_mu, "lambda": ret_lam, "psi": ret_psi, "nu": ret_nu}


def gumbel_max_sampling(logp, size=None, *, rng=None):
    if size is None:
        ret = np.argmax(logp -
                        np.log(-np.log(rng.uniform(size=logp.shape))), axis=0)
    else:
        ret = []
        for i in range(size):
            ret.append(np.argmax(logp -
                                 np.log(-np.log(rng.uniform(size=len(logp))))))
        ret = np.array(ret)
    return ret


def rejection_sample(f, max_y, a=0, b=1, size=None, *, rng=None):
    if size is None:
        x = rng.uniform(a, b)
        y = rng.uniform(0, max_y)
        while y > f(x):
            x = rng.uniform(a, b)
            y = rng.uniform(0, max_y)
        return x
    else:
        x = rng.uniform(a, b, size)
        y = rng.uniform(0, max_y, size)
        while np.any(y > f(x)):
            x[y > f(x)] = rng.uniform(a, b, np.sum(y > f(x)))
            y[y > f(x)] = rng.uniform(0, max_y, np.sum(y > f(x)))
        return x


def mixture_density(x, w, mu, sigma, u):
    k = len(w)

    ret = []
    for j in range(k):
        ret.append(multivariate_normal.pdf(x,
                                           mu[j],
                                           sigma[j],
                                           1))

    ret = np.array(ret).T
    mask = (np.array(list(repeat(u, k))) <
            np.array(list(repeat(w, len(u)))).transpose())

    ret = np.atleast_2d(ret.dot(mask / mask.sum(0))).mean(1)
    return ret


def cluster(x, w, mu, sigma, u):
    k = len(w)
    ret = []
    for j in range(k):
        ret.append(multivariate_normal.pdf(x,
                                           mu[j],
                                           sigma[j],
                                           1))
    ret = np.array(ret).T
    mask = (np.array(list(repeat(u, k))) <
            np.array(list(repeat(w, len(u)))).transpose())

    weights = (mask / mask.sum(0)).sum(1) / len(u)
    ret = ret * weights
    grp = np.argmax(ret, axis=1)
    u_grp, ret = np.unique(grp, return_inverse=True)
    return ret, weights[u_grp], mu[u_grp], sigma[u_grp]



from ._base import BaseGaussianMixture
from ..weight_models import GeometricProcess


class GeometricProcessMixture(BaseGaussianMixture):
    def __init__(self, *, a=1, b=1, rng=None, **kwargs):
        weight_model = GeometricProcess(a=a, b=b, rng=rng)
        super().__init__(weight_model=weight_model, rng=rng, **kwargs)

    def _get_run_params(self):
        return {"w": self.weight_model.get_weights(),
                "mu": self.mu,
                "sigma": self.sigma,
                "u": self.u,
                "d": self.d,
                "p": self.weight_model.get_p()}



from ._base import BaseWeights
import numpy as np
from scipy.stats import beta, binom
from scipy.optimize import minimize, brentq
from scipy.integrate import quad


class DGBProcess(BaseWeights):
    def __init__(self, p=0.5, theta=1, p_a=1, p_b=1, rng=None):
        super().__init__(rng=rng)
        self.p = p
        self.p_a = p_a
        self.p_b = p_b

        self.theta = theta

        self.v = np.array([], dtype=np.float64)
        self.bernoullis = np.array([], dtype=np.int)

    def structure_log_likelihood(self, v=None, bernoullis=None, p=None,
                                 theta=None):
        if v is None:
            v = self.v
        if bernoullis is None:
            bernoullis = self.bernoullis
        if p is None:
            p = self.p
        if theta is None:
            theta = self.theta
        log_likelihood = self.weight_log_likelihood(v=v, theta=theta)
        log_likelihood += self.persist_log_likelihood(bernoullis=bernoullis,
                                                      p=p)
        return log_likelihood

    def weight_log_likelihood(self, v=None, theta=None):
        if v is None:
            v = self.v
        if theta is None:
            theta = self.theta
        v = np.unique(v)
        log_likelihood = np.sum(beta.logpdf(v, a=1, b=theta))
        return log_likelihood

    def persist_log_likelihood(self, bernoullis=None, p=None):
        if bernoullis is None:
            bernoullis = self.bernoullis
        if p is None:
            p = self.p
        log_likelihood = binom.logpmf(k=np.sum(bernoullis),
                                      n=len(bernoullis),
                                      p=p)
        return log_likelihood

    def random(self, size=None):
        if size is None and len(self.d) == 0:
            raise ValueError("Weight structure not fitted and `n` not passed.")
        if size is not None:
            if type(size) is not int:
                raise TypeError("size parameter must be integer or None")
        if len(self.d) == 0:
            # self.random_p()
            mask_change = self.rng.binomial(n=1, p=1 - self.p, size=size - 1)
            mask_change = np.concatenate(([0], np.cumsum(mask_change)))
            self.v = self.rng.beta(a=1, b=self.theta, size=mask_change[-1] + 1)
            self.v = self.v[mask_change]
            self.w = self.v * np.cumprod(np.concatenate(([1],
                                                         1 - self.v[:-1])))
        else:
            self.random_p()
            self.random_bernoullis()
            remap = np.cumsum(self.bernoullis)
            a_c = np.bincount(self.d)
            b_c = np.concatenate((np.cumsum(a_c[::-1])[-2::-1], [0]))

            a_c = np.bincount(remap, a_c)
            b_c = np.bincount(remap, b_c)

            self.v = self.rng.beta(a=1 + a_c, b=self.theta + b_c)
            self.v = self.v[remap]
            self.w = self.v * np.cumprod(np.concatenate(([1],
                                                         1 - self.v[:-1])))
            if size is not None:
                self.complete(size)
        return self.w

    def complete(self, size):
        if type(size) is not int:
            raise TypeError("size parameter must be integer or None")
        if len(self.v) < size:
            if len(self.v) == 0:
                self.v = self.rng.beta(a=1, b=self.theta, size=1)
            mask_change = self.rng.binomial(n=1,
                                            p=1 - self.p,
                                            size=size - len(self.v))
            self.bernoullis = np.concatenate((self.bernoullis, mask_change))
            mask_change = np.cumsum(mask_change)
            temp_v = np.concatenate((
                [self.v[-1]],
                self.rng.beta(a=1, b=self.theta, size=mask_change[-1])))
            self.v = np.concatenate((self.v, temp_v[mask_change]))
            self.w = self.v * np.cumprod(np.concatenate(([1],
                                                         1 - self.v[:-1])))
        return self.w

    def tail(self, x):
        if x >= 1 or x < 0:
            raise ValueError("Tail parameter not in range [0,1)")
        if len(self.w) == 0:
            self.random(1)

        w_sum = sum(self.w)
        while w_sum < x:
            bool_new_val = self.rng.binomial(n=1, p=1 - self.p)
            self.bernoullis = np.concatenate((self.bernoullis, [bool_new_val]))
            if bool_new_val:
                v_to_append = self.rng.beta(a=1, b=self.theta, size=1)
                self.v = np.concatenate((self.v, v_to_append))
            else:
                self.v = np.concatenate((self.v, [self.v[-1]]))
            self.w = np.concatenate((self.w, [(1 - sum(self.w)) * self.v[-1]]))
            w_sum += self.w[-1]
        return self.w

    def random_bernoullis(self):
        temp_bernoullis = self.rng.binomial(n=1, p=1 - self.p, size=max(self.d))
        self.bernoullis = np.concatenate(([0], temp_bernoullis))
        return self.bernoullis

    def random_p(self):
        self.p = self.rng.beta(a=self.p_a + sum(self.bernoullis),
                               b=self.p_b + sum(1 - self.bernoullis))
        return self.p

    def get_p(self):
        return self.p



__all__ = []

from ._beta_bernoulli_mixture import BetaBernoulliMixture
from ._beta_binomial_mixture import BetaBinomialMixture
from ._beta_in_beta_mixture import BetaInBetaMixture
from ._dirichlet_distribution_mixture import DirichletDistributionMixture
from ._dirichlet_process_mixture import DirichletProcessMixture
from ._geometric_process_mixture import GeometricProcessMixture
from ._pitman_yor_process_mixture import PitmanYorMixture



from ._base import BaseWeights
import numpy as np
from scipy.stats import dirichlet


class DirichletDistribution(BaseWeights):
    def __init__(self, n=1, alpha=1, rng=None):
        super().__init__(rng=rng)
        assert type(n) == int, "parameter n must be of type int"
        self.n = n
        if type(alpha) in (list, np.ndarray):
            self.n = n
            self.alpha = alpha
        elif type(alpha) in (int, float):
            self.alpha = [alpha] * self.n
        self.v = np.array([], dtype=np.float64)

    def structure_log_likelihood(self, v=None, alpha=None):
        if v is None:
            v = self.v
        if alpha is None:
            alpha = self.alpha
        return self.weight_log_likelihood(v=v, theta=alpha)

    def weight_log_likelihood(self, v=None, theta=None):
        if v is None:
            v = self.v
        if theta is None:
            theta = self.alpha
        if len(v) == 0:
            return 0
        return np.sum(dirichlet.logpdf(v, theta))

    def random(self, size=None):
        if len(self.d) > 0:
            if max(self.d) >= len(self.alpha):
                raise ValueError(
                    'fitted structure is incompatible with this model'
                )
            else:
                a_c = np.bincount(self.d)
                a_c.resize(len(self.alpha), refcheck=False)
                self.v = self.rng.dirichlet(self.alpha + a_c)
        else:
            self.v = self.rng.dirichlet(self.alpha)
        self.w = self.v * np.cumprod(np.concatenate(([1],
                                                     1 - self.v[:-1])))
        return self.w

    def complete(self, size=None):
        if len(self.w) == 0:
            self.random(None)
        return self.w

    def tail(self, x=None):
        if len(self.w) == 0:
            self.random(None)
        return self.w



from sklearn.cluster import SpectralClustering
from scipy.stats import multivariate_normal
import pandas as pd
import numpy as np

from abc import ABCMeta
from . import _utils


class BaseGaussianMixture(metaclass=ABCMeta):
    def __init__(self, weight_model=None, mu_prior=None, lambda_prior=1,
                 psi_prior=None, nu_prior=None, total_iter=1000, burn_in=100,
                 subsample_steps=1, show_progress=False, rng=None):
        if rng is None:
            self.rng = np.random.default_rng()
        elif type(rng) is int:
            self.rng = np.random.default_rng(rng)
        else:
            self.rng = rng

        assert total_iter > burn_in, (
            "total_iter must be greater than burn_in period")
        self.burn_in = int(burn_in)
        self.total_iter = int(total_iter)
        self.subsample_steps = int(subsample_steps)

        self.mu_prior = mu_prior
        self.lambda_prior = lambda_prior
        self.psi_prior = psi_prior
        self.nu_prior = nu_prior

        self.weight_model = weight_model

        self.y = np.array([])
        self.d = np.array([])
        self.mu = np.array([])
        self.sigma = np.array([[]])
        self.u = np.array([])

        self.affinity_matrix = np.array([])

        self.map_sim_params = None
        self.map_log_likelihood = -np.inf
        self.total_saved_steps = 0
        self.sim_params = []
        self.n_groups = []
        self.n_atoms = []
        self.n_log_likelihood = []
        self.show_progress = show_progress

    def fit_gibbs(self, y, warm_start=False):
        if isinstance(y, pd.DataFrame):
            self.y = y.to_numpy()
        elif isinstance(y, list):
            self.y = np.array(y)
        elif isinstance(y, np.ndarray):
            self.y = y
        else:
            raise TypeError('type is not valid')

        if self.mu_prior is None:
            self.mu_prior = self.y.mean(axis=0)
        if self.psi_prior is None:
            self.psi_prior = np.atleast_2d(np.cov(self.y.T))
        if self.nu_prior is None:
            _, self.nu_prior = self.y.shape
        self.mu = self.mu.reshape(0, *self.mu_prior.shape)
        self.sigma = self.sigma.reshape(0, *self.psi_prior.shape)

        if not warm_start:
            self.sim_params = []
            self.n_groups = []
            self.n_atoms = []
            self.total_saved_steps = 0

            self.affinity_matrix = np.zeros((len(self.y), len(self.y)))

            self.u = self.rng.uniform(0 + np.finfo(np.float64).eps, 1,
                                      len(self.y))
            self.weight_model.tail(1 - min(self.u))

            self.d = self.weight_model.random_assignment(len(self.y))

            self._complete_atoms()
        self._train_gibbs()

    def fit_em(self, y, n=10, warm_start=False):
        if isinstance(y, pd.DataFrame):
            self.y = y.to_numpy()
        else:
            self.y = y

        if self.mu_prior is None:
            self.mu_prior = self.y.mean(axis=0)
        if self.psi_prior is None:
            self.psi_prior = np.atleast_2d(np.cov(self.y.T))
        if self.nu_prior is None:
            _, self.nu_prior = self.y.shape
        self.mu = self.mu.reshape(0, *self.mu_prior.shape)
        self.sigma = self.sigma.reshape(0, *self.psi_prior.shape)

        if not warm_start:
            self.sim_params = []
            self.n_groups = []
            self.n_atoms = []
            self.total_saved_steps = 0

            self.affinity_matrix = np.zeros((len(self.y), len(self.y)))

            self.u = self.rng.uniform(0 + np.finfo(np.float64).eps, 1, len(self.y))
            self.mu, self.sigma = _utils.random_normal_invw(
                mu=self.mu_prior,
                lam=self.lambda_prior,
                psi=self.psi_prior,
                nu=self.nu_prior,
                rng=self.rng)
            self.mu = np.array([self.mu])
            self.sigma = np.array([self.sigma])

            self.weight_model.tail(1 - min(self.u))
            self.d = self.weight_model.random_assignment(len(self.y))
            self._complete_atoms()
        self._train_em()

    def _update_weights(self):
        self.weight_model.fit(self.d)
        w = self.weight_model.random()
        self.u = self.rng.uniform(0 + np.finfo(np.float64).eps,
                                  w[self.d] + np.finfo(np.float64).eps)
        self.weight_model.tail(1 - min(self.u))

    def gibbs_eap_density(self, y, periods=None):
        y_sim = []
        if periods is None:
            for param in self.sim_params:
                y_sim.append(_utils.mixture_density(y,
                                                    param["w"],
                                                    param["mu"],
                                                    param["sigma"],
                                                    param["u"]))
        else:
            periods = min(periods, len(self.sim_params))
            for param in self.sim_params[-periods:]:
                y_sim.append(_utils.mixture_density(y,
                                                    param["w"],
                                                    param["mu"],
                                                    param["sigma"],
                                                    param["u"]))
        return np.array(y_sim).mean(axis=0)

    def gibbs_eap_affinity_matrix(self, y=None):
        if y is None:
            return self.affinity_matrix / self.total_saved_steps
        affinity_matrix = np.zeros((len(y), len(y)))
        for params in self.sim_params:
            grouping = _utils.cluster(y,
                                      params["w"],
                                      params["mu"],
                                      params["sigma"],
                                      params["u"])
            affinity_matrix += np.equal(grouping, grouping[:, None])
        affinity_matrix /= len(self.sim_params)
        return affinity_matrix

    def gibbs_eap_spectral_consensus_cluster(self, y=None, n_clusters=1):
        sc = SpectralClustering(n_clusters=n_clusters, affinity='precomputed')
        return sc.fit_predict(self.gibbs_eap_affinity_matrix(y))

    def gibbs_map_density(self, y=None):
        if y is None:
            y = self.y
        return _utils.mixture_density(y,
                                      self.map_sim_params["w"],
                                      self.map_sim_params["mu"],
                                      self.map_sim_params["sigma"],
                                      self.map_sim_params["u"])

    def gibbs_map_cluster(self, y=None, full=False):
        if y is None:
            y = self.y
        ret = _utils.cluster(y,
                             self.map_sim_params["w"],
                             self.map_sim_params["mu"],
                             self.map_sim_params["sigma"],
                             self.map_sim_params["u"])
        if full:
            return ret
        return ret[0]

    def get_n_groups(self):
        return self.n_groups

    def get_n_theta(self):
        return self.n_atoms

    def get_sim_params(self):
        return self.sim_params

    def _get_run_params(self):
        return {"w": self.weight_model.get_weights(),
                "mu": self.mu,
                "sigma": self.sigma,
                "u": self.u,
                "d": self.d}

    def _save_params(self):
        self.sim_params.append(self._get_run_params())
        self.n_groups.append(len(np.unique(self.d)))
        self.n_atoms.append(len(self.mu))
        self._update_map_params()
        self.affinity_matrix += np.equal(self.d, self.d[:, None])
        self.total_saved_steps += 1

    def _update_map_params(self):
        run_log_likelihood = self._full_log_likelihood()
        if self.map_log_likelihood < run_log_likelihood:
            self.map_log_likelihood = run_log_likelihood
            self.map_sim_params = self._get_run_params()
        elif self.map_log_likelihood == -np.inf:
            # It's better than nothing
            self.map_sim_params = self._get_run_params()

    def _update_atoms(self):
        assert len(self.mu) == len(self.sigma)
        self.d = np.unique(self.d, return_inverse=True)[1]
        self.mu = []
        self.sigma = []
        for j in range(max(self.d) + 1):
            inj = (self.d == j).nonzero()[0]

            posterior_params = _utils.posterior_norm_invw_params(
                self.y[inj],
                mu=self.mu_prior,
                lam=self.lambda_prior,
                psi=self.psi_prior,
                nu=self.nu_prior)
            temp_mu, temp_sigma = _utils.random_normal_invw(
                mu=posterior_params["mu"],
                lam=posterior_params["lambda"],
                psi=posterior_params["psi"],
                nu=posterior_params["nu"],
                rng=self.rng)
            self.mu.append(temp_mu)
            self.sigma.append(temp_sigma)

        self.mu = np.array(self.mu).reshape(-1, *self.mu_prior.shape)
        self.sigma = np.array(self.sigma).reshape(-1, *self.psi_prior.shape)

    def _complete_atoms(self):
        missing_len = self.weight_model.get_size() - len(self.mu)
        for _ in range(missing_len):
            temp_mu, temp_sigma = _utils.random_normal_invw(
                mu=self.mu_prior,
                lam=self.lambda_prior,
                psi=self.psi_prior,
                nu=self.nu_prior,
                rng=self.rng
            )
            self.mu = np.concatenate((self.mu, [np.atleast_1d(temp_mu)]))
            self.sigma = np.concatenate((self.sigma,
                                         [np.atleast_2d(temp_sigma)]))

    def _update_d(self):
        logproba = self._d_log_likelihood_vector()
        self.d = _utils.gumbel_max_sampling(logproba, rng=self.rng)

    def _train_gibbs(self):
        if self.show_progress:
            print("Starting burn-in.")
            from tqdm import trange
            for _ in trange(self.burn_in):
                self._gibbs_step()
            print("Finished burn-in.")
            print("Starting training.")
            for i in trange(self.total_iter - self.burn_in):
                self._gibbs_step()
                if i % self.subsample_steps == 0:
                    self._save_params()
            print("Finished training.")
        else:
            for _ in range(self.burn_in):
                self._gibbs_step()
            for i in range(self.total_iter - self.burn_in):
                self._gibbs_step()
                if i % self.subsample_steps == 0:
                    self._save_params()

    def _gibbs_step(self):
        self._update_atoms()
        self._update_weights()
        self._complete_atoms()
        self._update_d()

    def _train_em(self):
        if self.show_progress:
            print("Starting burn-in.")
            from tqdm import trange
            for _ in trange(self.burn_in):
                self._e_step()
                self._m_step()
            print("Finished burn-in.")
            print("Starting training.")
            for i in trange(self.total_iter - self.burn_in):
                self._e_step()
                self._m_step()
                if i % self.subsample_steps == 0:
                    self._save_params()
            print("Finished training.")
        else:
            for _ in range(self.burn_in):
                self._e_step()
                self._m_step()
            for i in range(self.total_iter - self.burn_in):
                self._e_step()
                self._m_step()
                if i % self.subsample_steps == 0:
                    self._save_params()

    def _d_log_likelihood_vector(self):
        with np.errstate(divide='ignore'):
            logproba = np.array([multivariate_normal.logpdf(self.y,
                                                            self.mu[j],
                                                            self.sigma[j],
                                                            1)
                                 for j in range(self.weight_model.get_size())])
            logproba += np.log(np.greater.outer(self.weight_model.get_weights(),
                                                self.u))
        return logproba

    def _y_log_likelihood(self):
        ret = 0
        with np.errstate(divide='ignore'):
            for di in np.unique(self.d):
                ret += np.sum(multivariate_normal.logpdf(self.y[self.d == di],
                                                         self.mu[di],
                                                         self.sigma[di],
                                                         1))
        return ret

    def _mixture_log_likelihood(self):
        ret_log_likelihood = self._y_log_likelihood()
        # Bellow is commented out since d's distribution is discrete uniform
        # ret_log_likelihood += np.sum(self._d_log_likelihood_vector()[self.d])
        return ret_log_likelihood

    def _full_log_likelihood(self):
        ret = self._mixture_log_likelihood()
        ret += self.weight_model.structure_log_likelihood()
        return ret



from ._base import BaseGaussianMixture
from ..weight_models import BetaBinomial


class BetaBinomialMixture(BaseGaussianMixture):
    def __init__(self, *, p=0.5, theta=1, mu_prior=None, lambda_prior=1,
                 psi_prior=None, nu_prior=None, total_iter=1000, burn_in=100,
                 subsample_steps=1, rng=None):
        weight_model = BetaBinomial(p=p, alpha=theta, rng=rng)
        super().__init__(weight_model=weight_model, mu_prior=mu_prior,
                         lambda_prior=lambda_prior, psi_prior=psi_prior,
                         nu_prior=nu_prior, total_iter=total_iter,
                         burn_in=burn_in, subsample_steps=subsample_steps,
                         rng=rng)



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pyrichlet",
    version="0.0.1",
    author="Fidel Selva",
    author_email="cfso100@gmail.com",
    description="A package for density estimation and clustering using "
                "infinite gaussian mixtures with stick-breaking weighting"
                "structures",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/cabo40/pyrichlet",
    project_urls={
        'Bug Tracker': 'https://github.com/cabo40/pyrichlet/issues',
        'Documentation': 'https://pyrichlet.readthedocs.io',
        'Source Code': 'https://github.com/cabo40/pyrichlet'
    },
    license='GPL3',
    packages=setuptools.find_packages(exclude=['tests']),
    install_requires=[
        'numpy',
        'scipy',
        'pandas',
        'scikit-learn'
    ],
    extras_require={
        "tqdm": ["tqdm"],
    },
    classifiers=[
        "Intended Audience :: Science/Research",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



from ._base import BaseWeights
import numpy as np
from scipy.stats import beta
from scipy.optimize import minimize, brentq
from scipy.integrate import quad


class BetaInBeta(BaseWeights):
    def __init__(self, x=0.5, alpha=1, a=1, b=1, p=None, p_method="static",
                 p_optim_max_steps=10, rng=None):
        super().__init__(rng=rng)
        self.x = x
        self.a = a
        self.b = b
        self.alpha = alpha
        if p is None:
            self.p = self.rng.beta(a=self.a, b=self.b)
        else:
            self.p = p

        self.p_method = p_method
        self.v = np.array([], dtype=np.float64)
        self.p_optim_max_steps = p_optim_max_steps
        self._validate_params()

    def structure_log_likelihood(self, v=None, p=None, x=None, alpha=None):
        if v is None:
            v = self.v
        if p is None:
            p = self.p
        if x is None:
            x = self.x
        if alpha is None:
            alpha = self.alpha
        log_likelihood = self.weight_log_likelihood(v=v, p=p, x=x, alpha=alpha)
        log_likelihood += self.p_log_likelihood(p=p)
        return log_likelihood

    def weight_log_likelihood(self, v=None, p=None, x=None, alpha=None, a=None,
                              b=None):
        if v is None:
            v = self.v
        if p is None:
            p = self.p
        if x is None:
            x = self.x
        if alpha is None:
            alpha = self.alpha
        if a is None:
            alpha = self.alpha
        if b is None:
            alpha = self.alpha
        if x == 1:
            if len(v) == 0:
                return 0
            if np.all(v == v[0]):
                return 0
            else:
                return -np.inf
        return np.sum(
            beta.logpdf(v,
                        a=1 + x / (1 - x) * p,
                        b=alpha + x / (1 - x) * (1 - p)))

    def p_log_likelihood(self, p=None, a=None, b=None):
        if p is None:
            p = self.p
        if a is None:
            a = self.a
        if b is None:
            b = self.b
        return beta.logpdf(p, a=a, b=b)

    def random(self, size=None):
        if size is None and len(self.d) == 0:
            raise ValueError("Weight structure not fitted and `n` not passed.")
        if size is not None:
            if type(size) is not int:
                raise TypeError("size parameter must be integer or None")
        if len(self.d) == 0:
            self.p = self.rng.beta(a=self.a, b=self.b)
            self.v = self.rng.beta(
                a=1 + self.x / (1 - self.x) * self.p,
                b=self.alpha + self.x / (1 - self.x) * (1 - self.p),
                size=size
            )
            self.w = self.v * np.cumprod(np.concatenate(([1],
                                                         1 - self.v[:-1])))
        else:
            self.random_p()
            a_c = np.bincount(self.d)
            b_c = np.concatenate((np.cumsum(a_c[::-1])[-2::-1], [0]))

            if size is not None and size < len(a_c):
                a_c = a_c[:size]
                b_c = b_c[:size]

            self.v = self.rng.beta(
                a=1 + self.x / (1 - self.x) * self.p + a_c,
                b=self.alpha + self.x / (1 - self.x) * (1 - self.p) + b_c
            )
            self.w = self.v * np.cumprod(np.concatenate(([1],
                                                         1 - self.v[:-1])))
            if size is not None:
                self.complete(size)
        return self.w

    def complete(self, size):
        if type(size) is not int:
            raise TypeError("size parameter must be integer or None")
        if len(self.v) < size:
            self.v = np.concatenate(
                (self.v,
                 self.rng.beta(
                     a=1 + self.x / (1 - self.x) * self.p,
                     b=self.alpha + self.x / (1 - self.x) * (1 - self.p),
                     size=size - len(self.v)))
            )
            self.w = self.v * np.cumprod(np.concatenate(([1],
                                                         1 - self.v[:-1])))
        return self.w

    def tail(self, x):
        if x >= 1 or x < 0:
            raise ValueError("Tail parameter not in range [0,1)")
        if len(self.w) == 0:
            self.random(1)

        w_sum = sum(self.w)
        while w_sum < x:
            v_to_append = self.rng.beta(
                a=1 + self.x / (1 - self.x) * self.p,
                b=self.alpha + self.x / (1 - self.x) * (1 - self.p),
                size=1
            )
            self.v = np.concatenate((self.v, v_to_append))
            self.w = np.concatenate((self.w, [(1 - sum(self.w)) * self.v[-1]]))
            w_sum += self.w[-1]
        return self.w

    def random_p(self):
        if len(self.d) == 0:
            self.p = self.rng.beta(a=self.a, b=self.b)
            return self.p
        if self.x == 1:
            self.p = self.rng.beta(a=self.a + len(self.d),
                                   b=self.b + self.d.sum())
            return self.p
        elif self.x == 0:
            return self.p

        if self.p_method == "static":
            return self.p
        elif self.p_method == "independent":
            self.p = self.rng.beta(a=self.a, b=self.b)
            return self.p
        elif self.p_method == "geometric":
            self.p = self.rng.beta(a=self.a + len(self.d),
                                   b=self.b + self.d.sum())
            return self.p
        elif self.p_method == "rejection-sampling":
            prev_logp = self.structure_log_likelihood()
            curr_iter = 0
            while curr_iter < self.p_optim_max_steps:
                pass_var = self.rng.uniform(
                    low=0,
                    high=np.exp(self.weight_log_likelihood())
                )
                temp_p = self.rng.beta(a=self.a, b=self.b)
                curr_logp = self.structure_log_likelihood(p=temp_p)
                pass_condition = np.exp(curr_logp - prev_logp) > pass_var
                curr_iter += 1
                if pass_condition:
                    self.p = temp_p
                    return self.p
                prev_logp = curr_logp
            return self.p
        elif self.p_method == "max-likelihood":
            max_param = minimize(
                lambda p: -self.structure_log_likelihood(p=p),
                np.array([self.p]),
                bounds=[(0, 1)],
                options={'maxiter': self.p_optim_max_steps})
            if max_param.success:
                self.p = max_param.x[0]
            return self.p
        elif self.p_method == "inverse-sampling":
            unif = self.rng.uniform()

            def f(p):
                return np.exp(self.structure_log_likelihood(p=p))

            integral_normalization = quad(f, 0, 1)[0]

            def f_integral(p):
                return quad(f, 0, p)[0] / integral_normalization - unif

            try:
                self.p = brentq(f_integral, a=0, b=1)
            except ValueError:
                pass
            return self.p
        else:
            raise ValueError(f"unknown p-method")

    def update_method(self, p_method):
        self.p_method = p_method

    def get_p(self):
        return self.p

    def _validate_params(self):
        accepted_methods = ["static", "independent", "geometric",
                            "rejection-sampling", "max-likelihood",
                            "inverse-sampling"]
        if self.p_method not in accepted_methods:
            raise ValueError(f"p_method must be one of {accepted_methods}")
